/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/crc/crc.c                                              *
 * Created:       2002-03-11 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2002-2007 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: crc.c 436 2007-01-19 04:47:07Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>
#include <lib/crc.h>

#include "crc.h"


#define CRC_BUF_SIZE MISC_BUF_SIZE


static const char    *argv0 = NULL;

static int           par_enc = 0;
static int           par_dec = 0;

static int           par_binary = 0;
static int           par_pipe = 0;
static int           par_quiet = 0;
static unsigned      par_flags = 0;
static int           par_capital = 0;

static unsigned      par_size = 32;
static unsigned long par_poly = CRC_CRC32;


static
void prt_help (void)
{
	fputs (
		"Usage: crc [options] [input [output]]\n"
		"  -a, --algorithm string  Set a predefined algorithm\n"
		"  -b, --binary            Print CRC as binary [no]\n"
		"  -c, --capital           Use capital letters as hex digits [no]\n"
		"  -d, --decode            Remove CRC sum from input [no]\n"
		"  -e, --encode            Add CRC sum to output [no]\n"
		"  -g, --polynomial poly   Set generator polynomial [crc32]\n"
		"  -i string               Set input file name [stdin]\n"
		"  -m, --reverse-poly      Reverse the polynomial [no]\n"
		"  -n, --crc-bits int      Set CRC size [32]\n"
		"  -o string               Set output file name [stdout]\n"
		"  -p, --pipe              Pass input through to stdout [no]\n"
		"  -P, --posix             Use posix CRC algorithm [no]\n"
		"  -q, --quiet             Quiet operation [no]\n"
		"  -r, --reverse-input     Reverse the bit order in the input [no]\n"
		"  -R, --reverse-output    Reverse the bit order in the output [no]\n"
		"  -x, --invert-init       Invert the initial remainder [no]\n"
		"  -X, --invert-crc        Invert the CRC [no]\n",
		stdout
	);

	fflush (stdout);
}


static
char crc_get_hex (unsigned char val)
{
	if (val < 10) {
		return ((char) ('0' + val));
	}

	if (par_capital) {
		return ((char) ('A' + val - 10));
	}

	return ((char) ('a' + val - 10));
}

static
void crc_put_hex (FILE *fp, unsigned long val, unsigned bits)
{
	unsigned digit;

	while (bits >= 4) {
		digit = (val >> (bits - 4)) & 0x0f;
		digit = crc_get_hex (digit);
		fputc (digit, fp);
		bits -= 4;
	}
}

static
int do_crc12_enc (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned char  buf[CRC_BUF_SIZE + 2];
	unsigned short reg;
	crc12_t        crc;

	crc12_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc12_calc (&crc, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc12_done (&crc);

	reg = crc12_get_crc (&crc);

	buf[0] = (unsigned char) ((reg >> 8) & 0xff);
	buf[1] = (unsigned char) (reg & 0xff);

	if (fwrite (buf, 1, 2, out) != 2) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}

static
int do_crc12_dec (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned short reg;
	unsigned char  buf[CRC_BUF_SIZE + 2];
	crc12_t        crc;

	crc12_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE + 2, inp);

	while (n > 2) {
		crc12_calc (&crc, buf, n - 2);

		if (fwrite (buf, 1, n - 2, out) != (n - 2)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		buf[0] = buf[n - 2];
		buf[1] = buf[n - 1];

		n = 2 + fread (buf + 2, 1, CRC_BUF_SIZE, inp);
	}

	reg = buf[0] & 0xff;
	reg = (reg << 8) + (buf[1] & 0xff);

	crc12_done (&crc);

	if (reg != crc12_get_crc (&crc)) {
		if (par_quiet == 0) {
			fprintf (stderr, "%s: crc mismatch\n", argv0);
		}

		return (1);
	}

	return (0);
}

static
int do_crc12 (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned char  buf[CRC_BUF_SIZE + 4];
	unsigned short reg;
	crc12_t        crc;

	crc12_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc12_calc (&crc, buf, n);

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc12_done (&crc);

	reg = crc12_get_crc (&crc);

	if (par_binary) {
		buf[0] = (reg >> 8) & 0xff;
		buf[1] = reg & 0xff;
		if (fwrite (buf, 1, 2, out) != 2) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}
	}
	else {
		crc_put_hex (out, reg, 12);
		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}


static
int do_crc16_enc (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned char  buf[CRC_BUF_SIZE + 2];
	unsigned short reg;
	crc16_t        crc;

	crc16_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc16_calc (&crc, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc16_done (&crc);

	reg = crc16_get_crc (&crc);

	buf[0] = (unsigned char) ((reg >> 8) & 0xff);
	buf[1] = (unsigned char) (reg & 0xff);

	if (fwrite (buf, 1, 2, out) != 2) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}

static
int do_crc16_dec (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned short reg;
	unsigned char  buf[CRC_BUF_SIZE + 2];
	crc16_t        crc;

	crc16_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE + 2, inp);

	while (n > 2) {
		crc16_calc (&crc, buf, n - 2);

		if (fwrite (buf, 1, n - 2, out) != (n - 2)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		buf[0] = buf[n - 2];
		buf[1] = buf[n - 1];

		n = 2 + fread (buf + 2, 1, CRC_BUF_SIZE, inp);
	}

	reg = buf[0] & 0xff;
	reg = (reg << 8) + (buf[1] & 0xff);

	crc16_done (&crc);

	if (reg != crc16_get_crc (&crc)) {
		if (par_quiet == 0) {
			fprintf (stderr, "%s: crc mismatch\n", argv0);
		}

		return (1);
	}

	return (0);
}

static
int do_crc16 (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned char  buf[CRC_BUF_SIZE + 4];
	unsigned short reg;
	crc16_t        crc;

	crc16_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc16_calc (&crc, buf, n);

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc16_done (&crc);

	reg = crc16_get_crc (&crc);

	if (par_binary) {
		buf[0] = (reg >> 8) & 0xff;
		buf[1] = reg & 0xff;
		if (fwrite (buf, 1, 2, out) != 2) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}
	}
	else {
		crc_put_hex (out, reg, 16);
		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}


static
int do_crc32_enc (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned char buf[CRC_BUF_SIZE + 4];
	unsigned long reg;
	crc32_t       crc;

	crc32_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc32_calc (&crc, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc32_done (&crc);

	reg = crc32_get_crc (&crc);

	buf[0] = (unsigned char) ((reg >> 24) & 0xff);
	buf[1] = (unsigned char) ((reg >> 16) & 0xff);
	buf[2] = (unsigned char) ((reg >> 8) & 0xff);
	buf[3] = (unsigned char) (reg & 0xff);

	if (fwrite (buf, 1, 4, out) != 4) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}

static
int do_crc32_dec (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned long reg;
	unsigned char buf[CRC_BUF_SIZE + 4];
	crc32_t       crc;

	crc32_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE + 4, inp);

	while (n > 4) {
		crc32_calc (&crc, buf, n - 4);

		if (fwrite (buf, 1, n - 4, out) != (n - 4)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		buf[0] = buf[n - 4];
		buf[1] = buf[n - 3];
		buf[2] = buf[n - 2];
		buf[3] = buf[n - 1];

		n = 4 + fread (buf + 4, 1, CRC_BUF_SIZE, inp);
	}

	reg = buf[0] & 0xff;
	reg = (reg << 8) + (buf[1] & 0xff);
	reg = (reg << 8) + (buf[2] & 0xff);
	reg = (reg << 8) + (buf[3] & 0xff);

	crc32_done (&crc);

	if (reg != crc32_get_crc (&crc)) {
		if (par_quiet == 0) {
			fprintf (stderr, "%s: crc mismatch\n", argv0);
		}

		return (1);
	}

	return (0);
}

static
int do_crc32 (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned char buf[CRC_BUF_SIZE + 8];
	unsigned long reg;
	crc32_t       crc;

	crc32_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc32_calc (&crc, buf, n);

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc32_done (&crc);

	reg = crc32_get_crc (&crc);

	if (par_binary) {
		buf[0] = (reg >> 24) & 0xff;
		buf[1] = (reg >> 16) & 0xff;
		buf[2] = (reg >> 8) & 0xff;
		buf[3] = reg & 0xff;
		if (fwrite (buf, 1, 4, out) != 4) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}
	}
	else {
		crc_put_hex (out, reg, 32);

		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}

static
int crc_set_poly (const char *str)
{
	if (strcmp (str, "crc12") == 0) {
		par_size = 12;
		par_poly = CRC_CRC12;
	}
	else if (strcmp (str, "crc16") == 0) {
		par_size = 16;
		par_poly = CRC_CRC16;
	}
	else if (strcmp (str, "ccitt") == 0) {
		par_size = 16;
		par_poly = CRC_CRC16_CCITT;
	}
	else if (strcmp (str, "crc16-ccitt") == 0) {
		par_size = 16;
		par_poly = CRC_CRC16_CCITT;
	}
	else if (strcmp (str, "crc32") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32;
	}
	else if (strcmp (str, "crc32-c") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32_CRC32C;
	}
	else {
		if (str_get_ulng (str, &par_poly)) {
			return (1);
		}
	}

	return (0);
}

static
int crc_set_algorithm (const char *str)
{
	if (strcmp (str, "cdrom") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32_CDROM;
		par_flags = CRC_REVERSE_INP | CRC_REVERSE_OUT;
	}
	else if (strcmp (str, "crc32") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32;
		par_flags = 0;
	}
	else if (strcmp (str, "posix") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32;
		par_flags = CRC_POSIX;
	}
	else if (strcmp (str, "sfv") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32;
		par_flags = CRC_REVERSE_INP | CRC_REVERSE_OUT;
		par_flags |= CRC_INV_V0 | CRC_INV_V1;
	}
	else {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int        i, r;
	unsigned   fcnt;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("crc");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "a", "algorithm")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing algorithm\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (crc_set_algorithm (argv[i])) {
				fprintf (stderr, "%s: bad algorithm (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "b", "binary")) {
			par_binary = 1;
		}
		else if (str_isarg2 (argv[i], "c", "capital")) {
			par_capital = 1;
		}
		else if (str_isarg2 (argv[i], "d", "decode")) {
			par_enc = 0;
			par_dec = 1;
		}
		else if (str_isarg2 (argv[i], "e", "encode")) {
			par_enc = 1;
			par_dec = 0;
		}
		else if (str_isarg2 (argv[i], "g", "polynomial")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing CRC generator polynomial\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (crc_set_poly (argv[i])) {
				fprintf (stderr, "%s: bad polynomial (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr,
					"%s: missing input file name\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "l", "little-endian")) {
			par_flags |= CRC_REVERSE_INP;
		}
		else if (str_isarg2 (argv[i], "m", "reverse-poly")) {
			par_poly = crc_reverse_poly (par_poly, par_size);
		}
		else if (str_isarg2 (argv[i], "n", "crc-bits")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing CRC size\n",
					argv0
				);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_size)) {
				fprintf (stderr, "%s: bad CRC size (%s)\n",
					argv0, argv[i]
				);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "p", "pipe")) {
			par_pipe = 1;
		}
		else if (str_isarg2 (argv[i], "P", "posix")) {
			par_flags |= CRC_POSIX;
		}
		else if (str_isarg2 (argv[i], "q", "quiet")) {
			par_quiet = 1;
		}
		else if (str_isarg2 (argv[i], "r", "reverse-input")) {
			par_flags |= CRC_REVERSE_INP;
		}
		else if (str_isarg2 (argv[i], "R", "reverse-output")) {
			par_flags |= CRC_REVERSE_OUT;
		}
		else if (str_isarg2 (argv[i], "x", "invert-init")) {
			par_flags |= CRC_INV_V0;
		}
		else if (str_isarg2 (argv[i], "X", "invert-crc")) {
			par_flags |= CRC_INV_V1;
		}
		else if (str_isarg2 (argv[i], NULL, "crc12")) {
			par_size = 12;
			par_poly = CRC_CRC12;
		}
		else if (str_isarg2 (argv[i], NULL, "crc16")) {
			par_size = 16;
			par_poly = CRC_CRC16;
		}
		else if (str_isarg2 (argv[i], NULL, "crc32")) {
			par_size = 32;
			par_poly = CRC_CRC32;
		}
		else if (str_isarg2 (argv[i], NULL, "ccitt")) {
			par_size = 16;
			par_poly = CRC_CRC16_CCITT;
		}
		else if (str_isarg2 (argv[i], NULL, "cdrom")) {
			par_flags = CRC_REVERSE_INP | CRC_REVERSE_OUT;
			par_size = 32;
			par_poly = CRC_CRC32_CDROM;
		}
		else if (str_isarg2 (argv[i], NULL, "sfv")) {
			par_flags = CRC_REVERSE_INP | CRC_REVERSE_OUT;
			par_flags |= CRC_INV_V0 | CRC_INV_V1;
			par_size = 32;
			par_poly = CRC_CRC32;
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, "rb");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, (par_enc || par_dec) ? "wb" : "w");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
			return (EXIT_FAILURE);
		}
	}

	if (par_size == 12) {
		if (par_enc) {
			r = do_crc12_enc (inp, out);
		}
		else if (par_dec) {
			r = do_crc12_dec (inp, out);
		}
		else {
			r = do_crc12 (inp, out);
		}
	}
	else if (par_size == 16) {
		if (par_enc) {
			r = do_crc16_enc (inp, out);
		}
		else if (par_dec) {
			r = do_crc16_dec (inp, out);
		}
		else {
			r = do_crc16 (inp, out);
		}
	}
	else if (par_size == 32) {
		if (par_enc) {
			r = do_crc32_enc (inp, out);
		}
		else if (par_dec) {
			r = do_crc32_dec (inp, out);
		}
		else {
			r = do_crc32 (inp, out);
		}
	}
	else {
		fprintf (stderr, "%s: bad CRC size (%u)\n", argv0, par_size);
		r = 1;
	}

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
