/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/defer/defer.c                                          *
 * Created:       2004-06-30 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2006-07-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2004-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: defer.c 395 2006-07-07 20:25:32Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ioctl.h>

#include "defer.h"


#define DEFER_TMP_DIR "/tmp"


static const char    *argv0 = NULL;

static unsigned      par_interval = 10;

static unsigned      par_pause = 0;
static unsigned      par_delay = 0;

static int           par_max_delay_ok = 0;
static unsigned      par_max_delay = 0;

static int           par_max_load1_ok = 0;
static unsigned      par_max_load1 = 0;

static int           par_max_load2_ok = 0;
static unsigned      par_max_load2 = 0;

static int           par_max_load3_ok = 0;
static unsigned      par_max_load3 = 0;

static int           par_pid_ok = 0;
static unsigned long par_pid = 0;
static int           par_detach = 0;

static int           par_lock_ok = 0;
static int           par_lock_fd = -1;
static pid_t         par_lock_pid = 0;
static char          par_lock_name[256];
static int           par_lock_force = 0;

static int           par_have_sighup = 0;


static void defer_unlock (void);


static
void prt_help (void)
{
	fputs (
		"Usage: defer [options] [-c] command [args...]\n"
		"  -c, --command        Rest of command line is command\n"
		"  -d, --detach         Detach [no]\n"
		"  -i, --interval int   Set the test interval in seconds [10]\n"
		"  -k, --lock           Get a user specific lock [no]\n"
		"  -K, --force-lock     Force acquisition of lock [no]\n"
		"  -p, --pause int      Set the initial pause in seconds [0]\n"
		"  -P, --pid int        Set a pid to check for [none]\n"
		"  -t, --delay int      Set the delay in seconds [0]\n"
		"  -m, --max-delay int  Set the maximum delay [none]\n"
		"  -l, --max-load1 int  Set the maximum load average 1 [none]\n"
		"  -L, --max-load2 int  Set the maximum load average 2 [none]\n",
		stdout
	);

	fflush (stdout);
}

static
void sig_hup (int val)
{
	signal (SIGHUP, sig_hup);

	par_have_sighup = 1;
}

static
void sig_int (int val)
{
	signal (val, sig_int);

	if (par_lock_pid == 0) {
		/* no fork yet */
		defer_unlock();
		exit (1);
	}
}

static
void defer_detach (int retval)
{
	int   fd;
	pid_t pid;

	pid = fork();

	if (pid < 0) {
		perror ("fork");
		exit (1);
	}

	if (pid > 0) {
		exit (retval);
	}

	fd = open ("/dev/null", O_RDWR);
	if (fd < 0) {
		perror ("open: /dev/null:");
		exit (1);
	}

	if (isatty (0)) {
		close (0);
		dup2 (fd, 0);
	}

	if (isatty (1)) {
		close (1);
		dup2 (fd, 1);
	}

	if (isatty (2)) {
		close (2);
		dup2 (fd, 2);
	}

	close (fd);

#ifdef TIOCNOTTY
	if ((fd = open ("/dev/tty", O_RDWR)) >= 0) {
		ioctl (fd, TIOCNOTTY, 0);
		close (fd);
	}
#endif
}

static
int defer_lock (int force)
{
	char *tmp1, *tmp2;

	tmp1 = par_lock_name;

	tmp2 = getenv ("USER");

	if ((tmp2 != NULL) && (strlen (tmp2) < 128)) {
		sprintf (tmp1, DEFER_TMP_DIR "/defer-%s", tmp2);
	}
	else {
		sprintf (tmp1, DEFER_TMP_DIR "/defer-%u", (unsigned) geteuid());
	}

	if (mkdir (tmp1, 0700)) {
		if (errno != EEXIST) {
			return (0);
		}
	}

	strcat (tmp1, "/lock");

	if (force) {
		unlink (tmp1);
	}

	par_lock_fd = open (tmp1, O_CREAT | O_EXCL, S_IWUSR | S_IRUSR);
	if (par_lock_fd == -1) {
		return (0);
	}

	return (1);
}

static
void defer_unlock (void)
{
	if (par_lock_fd >= 0) {
		unlink (par_lock_name);
		close (par_lock_fd);
		par_lock_fd = -1;
		par_lock_name[0] = 0;
	}
}

static
void defer_lock_wait (void)
{
	int   status;
	pid_t pid;
	pid_t ret;

	pid = fork();
	if (pid < 0) {
		perror ("fork");
		defer_unlock();
		exit (1);
	}

	if (pid == 0) {
		close (par_lock_fd);
		par_lock_fd = -1;
		return;
	}

	par_lock_pid = pid;

	while (1) {
		ret = wait (&status);

		if (ret == pid) {
			defer_unlock();
			exit (status);
		}
	}
}

static
int get_load (unsigned *v1, unsigned *v2, unsigned *v3)
{
	double f1, f2, f3;
	FILE   *fp;

	fp = fopen ("/proc/loadavg", "r");
	if (fp == NULL) {
		return (1);
	}

	if (fscanf (fp, "%lf %lf %lf", &f1, &f2, &f3) != 3) {
		fclose (fp);
		return (1);
	}

	fclose (fp);

	*v1 = (unsigned) (100.0 * f1);
	*v2 = (unsigned) (100.0 * f2);
	*v3 = (unsigned) (100.0 * f3);

	return (0);
}

static
int get_pid_ok (unsigned long pid)
{
	char fname[256];
	FILE *fp;

	sprintf (fname, "/proc/%lu/cmdline", pid);

	fp = fopen (fname, "rb");
	if (fp == NULL) {
		return (0);
	}

	fclose (fp);

	return (1);
}

static
int check_condition (void)
{
	if (par_have_sighup) {
		return (1);
	}

	if (par_delay > 0) {
		return (0);
	}

	if (par_max_load1_ok || par_max_load2_ok || par_max_load3_ok) {
		unsigned v1, v2, v3;

		if (get_load (&v1, &v2, &v3) == 0) {
			if (par_max_load1_ok && (v1 > par_max_load1)) {
				return (0);
			}
			if (par_max_load2_ok && (v2 > par_max_load2)) {
				return (0);
			}
			if (par_max_load3_ok && (v3 > par_max_load3)) {
				return (0);
			}
		}
	}

	if (par_pid_ok) {
		if (get_pid_ok (par_pid)) {
			return (0);
		}
	}

	if (par_lock_ok) {
		if (defer_lock (par_lock_force) == 0) {
			return (0);
		}
	}

	return (1);
}

static
int defer (int argc, char **argv)
{
	int      i;
	unsigned t, tt;
	char     **arg;

	tt = 0;

	while (check_condition() == 0) {
		if (par_max_delay_ok) {
			if (tt > par_max_delay) {
				break;
			}
		}

		t = par_interval;

		if (par_delay > 0) {
			if (par_delay < par_interval) {
				t = par_delay;
			}

			par_delay -= t;
		}

		tt += t;

		while ((par_have_sighup == 0) && (t > 0)) {
			t = sleep (t);
		}
	}

	if (par_lock_ok) {
		defer_lock_wait();
	}

	arg = malloc ((argc + 1) * sizeof (char **));
	if (arg == NULL) {
		return (1);
	}

	for (i = 0; i < argc; i++) {
		arg[i] = argv[i];
	}

	arg[argc] = NULL;

	execvp (arg[0], argv);

	return (1);
}

int main (int argc, char **argv)
{
	int i, r;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("defer");
			return (EXIT_SUCCESS);
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "d")) {
			par_detach = 1;
		}
		else if (str_isarg2 (argv[i], "i", "interval")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing interval time\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_interval)) {
				fprintf (stderr, "%s: bad interval time (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			if (par_interval == 0) {
				par_interval = 1;
			}
		}
		else if (str_isarg2 (argv[i], "t", "delay")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing delay time\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_delay)) {
				fprintf (stderr, "%s: bad delay time (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "k", "lock")) {
			par_lock_ok = 1;
		}
		else if (str_isarg2 (argv[i], "K", "force-lock")) {
			par_lock_ok = 1;
			par_lock_force = 1;
		}
		else if (str_isarg2 (argv[i], "p", "pause")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing pause time\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_pause)) {
				fprintf (stderr, "%s: bad pause time (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "P", "pid")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing pid\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_pid)) {
				fprintf (stderr, "%s: bad pid (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_pid_ok = 1;
		}
		else if (str_isarg2 (argv[i], "m", "max-delay")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing maximum delay\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_max_delay)) {
				fprintf (stderr, "%s: bad maximum delay (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_max_delay_ok = 1;
		}
		else if (str_isarg2 (argv[i], "l", "max-load1")) {
			double v;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing load\n", argv0);
				return (EXIT_FAILURE);
			}

			v = strtod (argv[i], NULL);

			par_max_load1_ok = 1;
			par_max_load1 = (unsigned) (100.0 * v);
		}
		else if (str_isarg2 (argv[i], "L", "max-load2")) {
			double v;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing load\n", argv0);
				return (EXIT_FAILURE);
			}

			v = strtod (argv[i], NULL);

			par_max_load2_ok = 1;
			par_max_load2 = (unsigned) (100.0 * v);
		}
		else if (str_isarg2 (argv[i], NULL, "max-load3")) {
			double v;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing load\n", argv0);
				return (EXIT_FAILURE);
			}

			v = strtod (argv[i], NULL);

			par_max_load3_ok = 1;
			par_max_load3 = (unsigned) (100.0 * v);
		}
		else if (str_isarg2 (argv[i], "c", "command")) {
			i += 1;
			break;
		}
		else if (argv[i][0] != '-') {
			break;
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (i >= argc) {
		return (EXIT_SUCCESS);
	}

	signal (SIGHUP, sig_hup);
	signal (SIGINT, sig_int);
	signal (SIGTERM, sig_int);

	while ((par_have_sighup == 0) && (par_pause > 0)) {
		par_pause = sleep (par_pause);
	}

	if (par_detach) {
		defer_detach (0);
	}

	r = defer (argc - i, argv + i);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
