/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     detty.c                                                    *
 * Created:       2002-09-10 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-07-11 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2002-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: detty.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/common.h>
#include <lib/file.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <stdio.h>
#include <stdlib.h>

#include "detty.h"


static const char *argv0 = NULL;

static int      par_trim = 0;
static unsigned par_width = 1024;


void prt_help (void)
{
	fputs (
		"Usage: detty [options] [input [output]]\n"
		"  --help           Print this help\n"
		"  --version        Print version information\n"
		"  -i string        Set input file name [stdin]\n"
		"  -o string        Set output file name [stdout]\n"
		"  -w, --width int  Set terminal width [1024]\n"
		"  -r, --trim       Remove spaces and tabs from end of lines [no]\n",
		stdout
	);

	fflush (stdout);
}

static
int is_space (unsigned char c)
{
	if ((c == ' ') || (c == '\t')) {
		return (1);
	}

	return (0);
}

static
void trim (unsigned char *buf, unsigned *n)
{
	while ((*n > 0) && is_space (buf[*n - 1])) {
		*n -= 1;
	}
}

static
int write_line (file_t *f_out, unsigned char *buf, unsigned n)
{
	if (par_trim) {
		trim (buf, &n);
	}

	if (n > 0) {
		if (file_write_block (f_out, buf, n) != n) {
			return (1);
		}
	}

	if (file_set_byte (f_out, '\n')) {
		return (1);
	}

	return (0);
}

/* this is far from perfect... */
static
int skip_escape (file_t *inp)
{
	unsigned char c;

	while (1) {
		if (file_get_byte (inp, &c)) {
			return (1);
		}

		if ((c == '[') || (c == ';')) {
			continue;
		}

		if ((c >= '0') && (c <= '9')) {
			continue;
		}

		return (0);
	}

	return (1);
}

static
int detty (file_t *f_inp, file_t *f_out)
{
	int           r;
	unsigned      i;
	unsigned char c;
	unsigned char *buf;
	unsigned      buf_i, buf_n;

	buf = (unsigned char *) malloc (par_width);
	buf_i = 0;
	buf_n = 0;

	r = file_get_byte (f_inp, &c);

	while (r == 0) {
		switch (c) {
			case 0x08:
				if (buf_i > 0) {
					for (i = buf_i; i < buf_n; i++) {
						buf[i - 1] = buf[i];
					}

					buf_i -= 1;
					buf_n -= 1;
				}
				break;

			case 0x0a:
				if (write_line (f_out, buf, buf_n)) {
					r = 1;
				}

				buf_i = 0;
				buf_n = 0;

				break;

			case 0x0d:
				buf_i = 0;
				break;

			case 0x1b:
				skip_escape (f_inp);
				break;

			default:
				if (buf_i < par_width) {
					buf[buf_i] = c;
					buf_i += 1;
					if (buf_i > buf_n) {
						buf_n = buf_i;
					}
				}
				break;
		}

		if (r == 0) {
			r = file_get_byte (f_inp, &c);
		}
	}

	if (buf_n > 0) {
		write_line (f_out, buf, buf_n);
	}

	free (buf);

	return (0);
}

int main (int argc, char *argv[])
{
	int      i, r;
	unsigned file_cnt;
	file_t   f_inp;
	file_t   f_out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("detty");
			return (EXIT_SUCCESS);
		}
	}

	file_set_stdin (&f_inp);
	file_set_stdout (&f_out);

	file_cnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			file_set_name (&f_inp, argv[i]);
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			file_set_name (&f_out, argv[i]);
		}
		else if (str_isarg2 (argv[i], "w", "width")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing width\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_width)) {
				fprintf (stderr, "%s: bad width (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			if (par_width == 0) {
				fprintf (stderr, "%s: width can't be 0\n", argv0);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "r", "rtrim")) {
			par_trim = 1;
		}
		else if (argv[i][0] != '-') {
			if (file_cnt == 0) {
				file_set_name (&f_inp, argv[i]);
				file_cnt = 1;
			}
			else if (file_cnt == 1) {
				file_set_name (&f_out, argv[i]);
				file_cnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (file_open_inp (&f_inp)) {
		fprintf (stderr, "%s: can't open input file\n", argv0);
		return (EXIT_FAILURE);
	}

	if (file_open_out (&f_out)) {
		fprintf (stderr, "%s: can't open output file\n", argv0);
		return (EXIT_FAILURE);
	}

	r = detty (&f_inp, &f_out);

	file_close (&f_out);
	file_close (&f_inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
