/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     entab.c                                                    *
 * Created:       2000-12-10 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2000-2007 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: entab.c 461 2007-10-15 21:25:26Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include "entab.h"


static const char *argv0 = NULL;

static int      par_entab = 1;
static int      par_trim = 0;
static int      par_beg = 0;
static unsigned par_min_step = 0;

static tab_lst  *par_tablst = NULL;


static
void prt_help (void)
{
	fputs (
		"entab: convert spaces to tabs and vice versa\n\n"
		"usage: entab [options] [input [output]]\n"
		"  -b, --begin            Restrict modifications to beginning of lines [no]\n"
		"  -d, --detab            Replace tabs by spaces\n"
		"  -e, --entab            Replace spaces by tabs [default]\n"
		"  -i string              Set input file name [stdin]\n"
		"  -m, --min-step int     Minimum number of spaces replaceable by a tab [1]\n"
		"  -o string              Set output file name [stdout]\n"
		"  -r, --trim             Remove spaces and tabs from end of lines [no]\n"
		"  -s, --default-tab int  Set default tabulator width [8]\n"
		"  -t, --tab [+]int       Set a (relative) tabulator\n",
		stdout
	);

	fflush (stdout);
}

static
int par_get_tab (const char *str, unsigned *val)
{
	unsigned        i;
	int             rel;
	static unsigned last = 0;

	rel = 0;

	i = 0;
	while (str[i] == ' ') {
		i += 1;
	}

	if (str[i] == '+') {
		rel = 1;
		i += 1;
	}

	if (str_get_uint (str + i, val)) {
		return (1);
	}

	if (rel) {
		*val += last;
	}

	last = *val;

	return (0);
}

static
line_buf *buf_new (unsigned n)
{
	line_buf *buf;

	if (n == 0) {
		n = 1;
	}

	buf = (line_buf *) malloc (sizeof (line_buf));
	if (buf == NULL) return (NULL);

	buf->chrcnt = 0;
	buf->chrmax = n;
	buf->chr = (char *) malloc (n);
	if (buf->chr == NULL) {
		free (buf);
		return (NULL);
	}

	buf->eolcnt = 0;

	return (buf);
}

static
void buf_free (line_buf *buf)
{
	if (buf != NULL) {
		free (buf->chr);
	}

	free (buf);
}

static
int buf_set_size (line_buf *buf, unsigned n)
{
	if (n < buf->chrcnt) {
		n = buf->chrcnt;
	}

	if (n == buf->chrmax) {
		return (0);
	}

	buf->chr = (char *) realloc (buf->chr, n);
	if (buf->chr == NULL) return (1);

	buf->chrmax = n;

	return (0);
}

static
void buf_add_char (line_buf *buf, char c)
{
	unsigned n;

	if (buf->chrcnt >= buf->chrmax) {
		n = 2 * buf->chrmax;
		if (n <= buf->chrcnt) return;

		buf_set_size (buf, n);
	}

	if (buf->chrcnt < buf->chrmax) {
		buf->chr[buf->chrcnt] = c;
		buf->chrcnt++;
	}
}

static
void buf_clear (line_buf *buf)
{
	buf->chrcnt = 0;
	buf->eolcnt = 0;

	if (buf->chrmax > 1024) {
		buf_set_size (buf, 1024);
	}
}

static
tab_lst *tab_new (void)
{
	tab_lst *ret;

	ret = (tab_lst *) malloc (sizeof (tab_lst));
	if (ret == NULL) return (NULL);

	ret->deftab = 8;
	ret->tab = NULL;

	return (ret);
}

static
void tab_free (tab_lst *lst)
{
	tab_lst_el *tab1, *tab2;

	if (lst != NULL) {
		tab1 = lst->tab;

		while (tab1 != NULL) {
			tab2 = tab1;
			tab1 = tab1->next;
			free (tab2);
		}

		free (lst);
	}
}

static
void tab_add (tab_lst *lst, unsigned pos)
{
	tab_lst_el *tab, *tmp;

	tab = (tab_lst_el *) malloc (sizeof (tab_lst_el));
	if (tab == NULL) return;

	tab->next = NULL;
	tab->pos = pos;

	if (lst->tab == NULL) {
		lst->tab = tab;
		return;
	}

	if (tab->pos < lst->tab->pos) {
		tab->next = lst->tab;
		lst->tab = tab;
		return;
	}

	tmp = lst->tab;
	while ((tmp->next != NULL) && (tmp->next->pos < tab->pos)) {
		tmp = tmp->next;
	}

	tab->next = tmp->next;
	tmp->next = tab;
}

static
unsigned tab_get_pos (tab_lst *lst, unsigned pos)
{
	tab_lst_el *tab;

	tab = lst->tab;
	while ((tab != NULL) && (tab->pos < pos)) {
		tab = tab->next;
	}

	if (tab != NULL) {
		return (tab->pos);
	}

	pos = pos + lst->deftab - 1;
	pos = pos / lst->deftab;
	pos = pos * lst->deftab;

	return (pos);
}

static
int ln_read (line_buf *buf, FILE *inp)
{
	int c;

	buf_clear (buf);

	c = fgetc (inp);
	if (c == EOF) {
		return (1);
	}

	while (c != EOF) {
		if (c == 0x0d) {
			buf->eol[0] = 0x0d;
			buf->eolcnt = 1;

			c = fgetc (inp);

			if (c != 0x0a) {
				if (c != EOF) {
					ungetc (c, inp);
				}
			}
			else {
				buf->eol[1] = 0x0a;
				buf->eolcnt = 2;
			}

			return (0);
		}
		else if (c == 0x0a) {
			buf->eol[0] = 0x0a;
			buf->eolcnt = 1;

			return (0);
		}
		else {
			buf_add_char (buf, c);

			c = fgetc (inp);
		}
	}

	return (0);
}

static
void ln_write (line_buf *buf, FILE *out)
{
	fwrite (buf->chr, 1, buf->chrcnt, out);
	fwrite (buf->eol, 1, buf->eolcnt, out);
}

static
void ln_rtrim (line_buf *buf)
{
	unsigned i;

	i = buf->chrcnt;

	while (i > 0) {
		i--;

		if ((buf->chr[i] == 0x20) || (buf->chr[i] == 0x09)) {
			buf->chrcnt--;
		}
		else {
			return;
		}
	}
}

static
void ln_entab (tab_lst *lst, line_buf *b_in, line_buf *b_out)
{
	unsigned i;
	char     c;
	unsigned col1, col2;
	unsigned tab;

	buf_clear (b_out);

	b_out->eolcnt = b_in->eolcnt;

	for (i = 0; i < b_in->eolcnt; i++) {
		b_out->eol[i] = b_in->eol[i];
	}

	col1 = b_out->chrcnt;
	col2 = b_out->chrcnt;

	for (i = 0; i < b_in->chrcnt; i++) {
		c = b_in->chr[i];

		switch (c) {
			case 0x20:
				col2++;
				break;

			case 0x09:
				col2 = tab_get_pos (lst, col2 + 1);
				break;

			default:
				if (col1 < col2) {
					tab = tab_get_pos (lst, col1 + 1);

					while ((tab <= col2) && ((col1 + par_min_step) < col2)) {
						buf_add_char (b_out, 0x09);
						col1 = tab;
						tab = tab_get_pos (lst, col1 + 1);
					}

					while (col1 < col2) {
						buf_add_char (b_out, 0x20);
						col1++;
					}
				}

				buf_add_char (b_out, (char)c);
				col1++;
				col2++;

				if (par_beg) {
					i++;

					while (i < b_in->chrcnt) {
						buf_add_char (b_out, b_in->chr[i]);
						i++;
					}

					return;
				}
				break;
		}
	}

	if (col1 < col2) {
		tab = tab_get_pos (lst, col1 + 1);

		while (tab <= col2) {
			buf_add_char (b_out, 0x09);
			col1 = tab;
			tab = tab_get_pos (lst, col1 + 1);
		}

		while (col1 < col2) {
			 buf_add_char (b_out, 0x20);
			 col1++;
		}
	}
}

static
void ln_detab (tab_lst *lst, line_buf *b_in, line_buf *b_out)
{
	unsigned i;
	char     c;
	unsigned col;
	unsigned tab;

	buf_clear (b_out);

	b_out->eolcnt = b_in->eolcnt;

	for (i = 0; i < b_in->eolcnt; i++) {
		b_out->eol[i] = b_in->eol[i];
	}

	col = b_out->chrcnt;

	for (i = 0; i < b_in->chrcnt; i++) {
		c = b_in->chr[i];

		switch (c) {
			case 0x09:
				tab = tab_get_pos (lst, col + 1 + par_min_step);

				while (col < tab) {
					buf_add_char (b_out, 0x20);
					col++;
				}
				break;

			case 0x20:
				buf_add_char (b_out, c);
				col++;
				break;

			default:
				buf_add_char (b_out, c);
				col++;

				if (par_beg) {
					i++;

					while (i < b_in->chrcnt) {
						buf_add_char (b_out, b_in->chr[i]);
						i++;
					}

					return;
				}
				break;
		}
	}
}

static
int entab (FILE *inp, FILE *out)
{
	int      r;
	line_buf *buf1, *buf2, *tmp;

	buf1 = buf_new (128);
	buf2 = buf_new (128);

	r = ln_read (buf1, inp);

	while (r == 0) {
		if (par_trim) {
			ln_rtrim (buf1);
		}

		if (par_entab) {
			ln_entab (par_tablst, buf1, buf2);
		}
		else {
			ln_detab (par_tablst, buf1, buf2);
		}

		tmp = buf1;
		buf1 = buf2;
		buf2 = tmp;

		ln_write (buf1, out);

		r = ln_read (buf1, inp);
	}

	buf_free (buf2);
	buf_free (buf1);

	return (0);
}

int main (int argc, char *argv[])
{
	int        i, r;
	unsigned   tab;
	unsigned   fcnt;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("entab");
			return (EXIT_SUCCESS);
		}
	}

	par_tablst = tab_new();
	if (par_tablst == NULL) {
		return (EXIT_FAILURE);
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "e", "entab")) {
			par_entab = 1;
		}
		else if (str_isarg2 (argv[i], "d", "detab")) {
			par_entab = 0;
		}
		else if (str_isarg2 (argv[i], "r", "trim")) {
			par_trim = 1;
		}
		else if (str_isarg2 (argv[i], "t", "tab")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing tab position\n", argv0);
				return (EXIT_FAILURE);
			}

			if (par_get_tab (argv[i], &tab)) {
				fprintf (stderr, "%s: bad tab position (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			tab_add (par_tablst, tab);
		}
		else if (str_isarg2 (argv[i], "s", "default-tab")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing default tab width\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_tablst->deftab)) {
				fprintf (stderr, "%s: bad default tab width (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			if (par_tablst->deftab == 0) {
				fprintf (stderr, "%s: tab width can't be 0\n", argv0);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "b", "begin")) {
			par_beg = 1;
		}
		else if (str_isarg2 (argv[i], "m", "min-step")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing minimum step size\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_min_step)) {
				fprintf (stderr, "%s: bad minimum step size (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			if (par_min_step > 0) {
				par_min_step -= 1;
			}
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, "r");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, "w");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
			return (EXIT_FAILURE);
		}
	}

	r = entab (inp, out);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	tab_free (par_tablst);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
