/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     inet.c                                                     *
 * Created:       2002-01-20 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-04-17 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2002-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: inet.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/common.h>
#include <lib/socket.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "inet.h"


static const char     *argv0 = NULL;

static int            par_listen = 1;
static int            par_verbose = 0;

static char           *par_host = NULL;
static char           *par_addr = NULL;
static unsigned short par_port = 0;

static unsigned       par_cnt = 0;
static unsigned       par_max = 1;
static unsigned long  par_total = 0;

static unsigned       par_delay = 1;

static char           *par_cmd = NULL;
static char           **par_arg = NULL;

static int            par_term = 0;

static int            par_fd;


static
void prt_help (void)
{
	fputs (
		"Usage: inet [options] -c command\n"
		"  --help                  Print usage information\n"
		"  --version               Print version information\n"
		"  -v, --verbose           Verbose operation\n"
		"  -i, --listen            Listen to a port [default]\n"
		"  -o, --connect           Connect to a port\n"
		"  -h, --host host[:port]  Set the host name\n"
		"  -p, --port int          Set the port\n"
		"  -m, --max int           Set maximum number of connections [1]\n"
		"  -n, --total int         Set total number of connections [0]\n"
		"  -r, --delay int         Set retry delay in seconds [1]\n"
		"  -c, --command string    Set command to execute\n",
		stdout
	);

	fflush (stdout);
}

static
void sig_term (int sig)
{
	par_term = 1;
}

static
void sig_pipe (int sig)
{
}

static
void inet_exec (int fd)
{
	close (0);
	dup2 (fd, 0);

	close (1);
	dup2 (fd, 1);

	close (fd);

	execvp (par_cmd, par_arg);

	close (0);
	close (1);

	exit (1);
}

static
int inet_fork (int fd)
{
	pid_t pid;

	if ((par_max > 0) && (par_cnt >= par_max)) {
		return (1);
	}

	pid = fork();

	if (pid < 0) {
		return (1);
	}
	else if (pid > 0) {
		par_cnt += 1;
		return (0);
	}

	close (par_fd);

	inet_exec (fd);

	exit (1);
}

static
int inet_wait (unsigned cnt)
{
	int r;
	int ret;

	while (1) {
		r = waitpid (-1, &ret, WNOHANG);

		if (r == 0) {
			break;
		}

		par_cnt -= 1;
	}

	while ((cnt > 0) && (par_cnt >= par_max) && (par_term == 0)) {
		r = wait (&ret);
		if (r > 0) {
			par_cnt -= 1;
		}
	}

	return (0);
}

static
int inet_listen (void)
{
	int            fd;
	char           *host;
	unsigned short port;

	par_fd = -1;

	while (par_fd < 0) {
		if (par_term) {
			return (1);
		}

		par_fd = tcp_listen (par_addr, par_port, 16);

		if (par_fd < 0) {
			if (par_delay == 0) {
				fprintf (stderr, "%s: can't create socket for port %u\n",
					argv0, par_port
				);
				return (1);
			}

			sleep (par_delay);
		}
	}

	while (par_term == 0) {
		fd = tcp_accept_time (par_fd, 1000);

		if (fd >= 0) {
			if (par_verbose) {
				if (tcp_get_peer (fd, &host, &port) == 0) {
					printf ("%s:%u -> %u\n", host, (unsigned) port, (unsigned) par_port);
					fflush (stdout);
					free (host);
				}
			}

			inet_fork (fd);

			close (fd);

			if (par_total > 0) {
				par_total -= 1;

				if (par_total == 0) {
					close (par_fd);

					inet_wait (1);

					return (0);
				}
			}

			inet_wait (par_max);
		}
	}

	close (par_fd);

	return (1);
}

static
int inet_connect (void)
{
	int fd;

	fd = -1;

	while (fd < 0) {
		if (par_term) {
			return (1);
		}

		fd = tcp_connect (par_addr, par_port);

		if (fd < 0) {
			if (par_delay == 0) {
				fprintf (stderr, "%s: connecting to %s:%hu failed\n",
					argv0, par_addr, par_port
				);
				return (1);
			}

			sleep (par_delay);
		}
	}

	inet_exec (fd);

	return (1);
}

int main (int argc, char **argv)
{
	int i, r;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("inet");
			return (EXIT_SUCCESS);
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "i", "listen")) {
			par_listen = 1;
		}
		else if (str_isarg2 (argv[i], "o", "connect")) {
			par_listen = 0;
		}
		else if (str_isarg2 (argv[i], "v", "verbose")) {
			par_verbose = 1;
		}
		else if (str_isarg2 (argv[i], "h", "host")) {
			unsigned short port;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing host name\n", argv0);
				return (EXIT_FAILURE);
			}

			par_host = tcp_addr_get_host (argv[i]);

			if (tcp_addr_get_port (argv[i], &port) == 0) {
				par_port = port;
			}
		}
		else if (str_isarg2 (argv[i], "p", "port")) {
			unsigned tmp;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing port number\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &tmp)) {
				fprintf (stderr, "%s: bad port number (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_port = tmp;
		}
		else if (str_isarg2 (argv[i], "m", "max")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing maximum connections\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_max)) {
				fprintf (stderr, "%s: bad connection count (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "n", "total")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing number of connections\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_total)) {
				fprintf (stderr, "%s: bad connection count (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "r", "delay")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing retry delay\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_delay)) {
				fprintf (stderr, "%s: bad delay (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "c", "command")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing command\n", argv0);
				return (EXIT_FAILURE);
			}

			par_cmd = argv[i];
			par_arg = &argv[i];

			i = argc;
		}
		else {
			fprintf (stderr, "%s: unknown parameter (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (par_cmd == NULL) {
		fprintf (stderr, "%s: need a command\n", argv0);
		return (EXIT_FAILURE);
	}

	if (par_host != NULL) {
		par_addr = tcp_host_resolve (par_host);
	}

	signal (SIGINT, &sig_term);
	signal (SIGPIPE, &sig_pipe);

	if (par_listen) {
		r = inet_listen();
	}
	else {
		r = inet_connect();
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
