/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lib/date.c                                             *
 * Created:       2005-07-04 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: date.c 414 2006-12-02 04:39:17Z hampa $ */


#include <stdio.h>

#include "date.h"


static
int tm_is_leap_year (unsigned year)
{
	if (((year % 4) == 0) && ((year % 100) != 0)) {
		return (1);
	}

	if ((year % 400) == 0) {
		return (1);
	}

	return (0);
}

static
unsigned tm_get_day_of_year (unsigned m, unsigned d, int leap)
{
	static unsigned days[2][12] = {
		{ 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 },
		{ 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335 }
	};

	if (m > 11) {
		return (0);
	}

	return (days[leap != 0][m] + d);
}

static
unsigned tm_get_month_length (unsigned m, int leap)
{
	static unsigned days[2][12] = {
		{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
		{ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
	};

	if (m > 11) {
		return (0);
	}

	return (days[leap != 0][m]);
}

unsigned tm_get_day_of_week (int64_t tm)
{
	unsigned v;

	if (tm < 0) {
		tm = (-tm + 86400LL * 1000LL - 1) / (86400LL * 1000LL);
		v = tm % 7;
		v = (12 - v) % 7;
	}
	else {
		tm = tm / (86400LL * 1000LL);
		v = (tm + 5) % 7;
	}

	return (v);
}

int tm_set_gregorian (int64_t *tm, int bc, unsigned y, unsigned m, unsigned d)
{
	unsigned long yy;
	unsigned      dd;
	int           leap;

	if ((y == 0) || (m < 1) || (m > 12)) {
		return (1);
	}

	if (bc) {
		y -= 1;
	}

	leap = tm_is_leap_year (y);

	if ((d < 1) || (d > tm_get_month_length (m - 1, leap))) {
		return (1);
	}

	if ((bc == 0) && (y >= 2000)) {
		yy = y - 2000;

		dd = (yy + 3) / 4;
		dd += (yy + 399) / 400;
		dd -= (yy + 99) / 100;

		*tm = 365 * (int64_t) yy + dd;
	}
	else {
		if (bc) {
			yy = 2000 + y;
		}
		else {
			yy = 2000 - y;
		}

		dd = yy / 4;
		dd += yy / 400;
		dd -= yy / 100;

		*tm = -(365 * (int64_t) yy + dd);
	}

	*tm += tm_get_day_of_year (m - 1, d - 1, leap);
	*tm *= 24;
	*tm *= 60 * 60;
	*tm *= 1000;

	return (0);
}

static
int tm_get_gregorian_1 (unsigned long days, unsigned *m, unsigned *d, int leap)
{
	unsigned i;
	unsigned *dc;

	static unsigned mdays[2][13] = {
		{ 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
		{ 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
	};

	dc = mdays[leap != 0];

	for (i = 0; i < 12; i++) {
		if (days < dc[i + 1]) {
			*m = i;
			*d = days - dc[i];
			return (0);
		}
	}

	return (1);
}

static
int tm_get_gregorian_4 (unsigned long days, long *y, unsigned *m, unsigned *d, int leap)
{
	if (leap) {
		if (days < 366) {
			return (tm_get_gregorian_1 (days, m, d, 1));
		}

		*y += (long) (1 + (days - 366) / 365);

		return (tm_get_gregorian_1 ((days - 366) % 365, m, d, 0));
	}
	else {
		*y += (long) (days / 365);

		return (tm_get_gregorian_1 (days % 365, m, d, 0));
	}
}

static
int tm_get_gregorian_100 (unsigned long days, long *y, unsigned *m, unsigned *d, int leap)
{
	if (leap == 0) {
		if (days < (4 * 365)) {
			return (tm_get_gregorian_4 (days, y, m, d, 0));
		}

		days -= (4 * 365);

		*y += 4L * (long) (1 + days / (4 * 365 + 1));

		return (tm_get_gregorian_4 (days % (4 * 365 + 1), y, m, d, 1));
	}
	else {
		*y += 4L * (long) (days / (4 * 365 + 1));

		return (tm_get_gregorian_4 (days % (4 * 365 + 1), y, m, d, 1));
	}
}

static
int tm_get_gregorian_400 (unsigned long days, long *y, unsigned *m, unsigned *d)
{
	if (days < (100UL * 365UL + 25)) {
		return (tm_get_gregorian_100 (days, y, m, d, 1));
	}
	else {
		days -= (100UL * 365UL + 25);

		*y += 100L * (long) (1 + days / (100UL * 365UL + 24));

		return (tm_get_gregorian_100 (days % (100UL * 365UL + 24), y, m, d, 0));
	}
}

int tm_get_gregorian (int64_t tm, int *bc, unsigned *y, unsigned *m, unsigned *d)
{
	unsigned long days;
	long          yy;

	if (tm < 0) {
		tm = (-tm + 86400UL * 1000UL - 1) / (86400UL * 1000UL);
		days = tm % (400 * 365 + 97);
		if (days > 0) {
			days = (400 * 365 + 97) - days;
		}
		yy = 2000L - (long) (400 * ((tm + (400 * 365 + 97) - 1) / (400 * 365 + 97)));
	}
	else {
		tm = tm / (86400UL * 1000UL);
		days = tm % (400 * 365 + 97);
		yy = 2000 + 400 * (tm / (400 * 365 + 97));
	}

	if (tm_get_gregorian_400 (days, &yy, m, d)) {
		return (1);
	}

	if (yy <= 0) {
		*y = -yy + 1;
		*bc = 1;
	}
	else {
		*y = yy;
		*bc = 0;
	}

	*m += 1;
	*d += 1;

	return (0);
}

int tm_set_julian (int64_t *tm, int bc, unsigned y, unsigned m, unsigned d)
{
	int64_t yy;
	int     leap;

	if (y == 0) {
		return (1);
	}

	if ((m < 1) || (m > 12)) {
		return (1);
	}

	if (bc) {
		y -= 1;
	}

	leap = ((y % 4) == 0);

	if ((d < 1) || (d > tm_get_month_length (m - 1, leap))) {
		return (1);
	}

	if (bc) {
		yy = 4712 - (int64_t) y;
	}
	else {
		yy = 4712 + y;
	}

	*tm = 365 * yy;

	if (yy < 0) {
		*tm += (-yy + 3) / 4;
	}
	else {
		*tm += (yy + 3) / 4;
	}

	*tm += tm_get_day_of_year (m - 1, d - 1, leap);
	*tm -= 2451545;
	*tm *= 24;
	*tm *= 60 * 60;
	*tm *= 1000;

	return (0);
}

int tm_get_julian (int64_t tm, int *bc, unsigned *y, unsigned *m, unsigned *d)
{
	unsigned long dd;
	long          yy;

	if (tm < 0) {
		tm = -((-tm + 86400UL * 1000UL - 1) / (86400UL * 1000UL));
	}
	else {
		tm = tm / (86400UL * 1000UL);
	}

	tm -= 13;

	if (tm < 0) {
		yy = 2000 - (long) (4 * ((-tm + 4 * 365) / (4 * 365 + 1)));
		dd = -tm % (4 * 365 + 1);
		if (dd > 0) {
			dd = (4 * 365 + 1) - dd;
		}
	}
	else {
		yy = 2000 + (long) (4 * (tm / (4 * 365 + 1)));
		dd = tm % (4 * 365 + 1);
	}

	if (tm_get_gregorian_4 (dd, &yy, m, d, 1)) {
		return (1);
	}

	if (yy <= 0) {
		*bc = 1;
		*y = -yy + 1;
	}
	else {
		*bc = 0;
		*y = yy;
	}

	*m += 1;
	*d += 1;

	return (0);
}

int tm_set_unix (int64_t *tm, int64_t val)
{
	*tm = 1000 * (val - 946684800);

	return (0);
}

int tm_get_unix (int64_t tm, int64_t *val)
{
	if (tm < 0) {
		tm = (-tm + 999) / 1000;
		*val = -tm + 946684800;
	}
	else {
		*val = (tm / 1000) + 946684800;
	}

	return (0);
}

int tm_set_jd (int64_t *tm, double jd)
{
	jd -= 2451544.5;

	*tm = (int64_t) (86400.0 * 1000.0 * jd);

	return (0);
}

int tm_get_jd (int64_t tm, double *jd)
{
	*jd = tm / (86400.0 * 1000.0);
	*jd += 2451544.5;

	return (0);
}

int tm_set_time (int64_t *tm, unsigned h, unsigned m, unsigned s, unsigned ms)
{
	int64_t val;

	if (*tm < 0) {
		*tm = -*tm + 86400UL * 1000UL - 1;
		*tm -= (*tm % (86400UL * 1000UL));
		*tm = -*tm;
	}
	else {
		*tm -= (*tm % (86400UL * 1000UL));
	}

	val = 60 * h + m;
	val = 60 * val + s;
	val = 1000 * val + ms;

	*tm += val;

	return (0);
}

int tm_get_time (int64_t tm, unsigned *h, unsigned *m, unsigned *s, unsigned *ms)
{
	unsigned long val;

	if (tm < 0) {
		val = (unsigned long) (-tm % (86400UL * 1000UL));
		if (val > 0) {
			val = 86400UL * 1000UL - val;
		}
	}
	else {
		val = (unsigned long) (tm % (86400UL * 1000UL));
	}

	*ms = val % 1000;
	val = val / 1000;

	*s = val % 60;
	val = val / 60;

	*m = val % 60;
	*h = val / 60;

	return (0);
}

int tm_add_time (int64_t *tm, unsigned h, unsigned m, unsigned s, unsigned ms)
{
	int64_t val;

	val = 60 * h + m;
	val = 60 * val + s;
	val = 1000 * val + ms;

	*tm += val;

	return (0);
}
