/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/relink/relink.c                                        *
 * Created:       2008-02-19 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2008 Hampa Hug <hampa@hampa.ch>                        *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: relink.c 477 2008-07-15 08:01:51Z hampa $ */


#include <lib/common.h>
#include <lib/crc.h>
#include <lib/strarg.h>

#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "relink.h"


#define RELINK_BUF_SIZE MISC_BUF_SIZE


typedef struct {
	char               *name;
	unsigned long long size;
	unsigned long      chk;
	unsigned           set;
	dev_t              dev;
	ino_t              ino;
	mode_t             mode;
	time_t             mtime;
	uid_t              uid;
	gid_t              gid;
	char               done;
} file_entry_t;


typedef struct {
	unsigned     cnt;
	unsigned     max;
	file_entry_t **ent;
} file_group_t;


typedef struct {
	unsigned long cnt;
	unsigned long max;
	file_entry_t  *ent;
} file_list_t;


static const char    *argv0 = NULL;

static int           par_verbose = 0;
static int           par_progress = 0;
static int           par_reverse = 0;
static int           par_use_mode = 1;
static int           par_use_uid = 1;
static int           par_use_gid = 1;
static int           par_use_mtime = 1;
static int           par_link = 0;
static int           par_print_same = 0;
static int           par_set_diff = 0;

static unsigned char par_buf[RELINK_BUF_SIZE];

static unsigned long par_chk_cnt = 0;
static unsigned long par_cmp_cnt = 0;
static unsigned long par_lnk_cnt = 0;

static crc32_t       par_crc;


static
void prt_help (void)
{
	fputs (
		"relink: find identical files and create links\n"
		"\n"
		"usage: relink [options] [input]\n"
		"  -g, --no-gid           Don't compare gid [no]\n"
		"  -i, --input string     Add a file list [stdin]\n"
		"  -l, --link             Relink files [no]\n"
		"  -m, --no-mode          Don't compare mode [no]\n"
		"  -n, --next-set string  Add a new set\n"
		"  -p, --progress         Print progress to stderr [no]\n"
		"  -r, --reverse          Reverse sort order [no]\n"
		"  -s, --print-same       Print existing links [no]\n"
		"  -t, --no-mtime         Don't compare mtime [no]\n"
		"  -u, --no-uid           Don't compare uid [no]\n"
		"  -v, --verbose          Verbose operation [no]\n",
		stdout
	);

	fflush (stdout);
}



static
void relink_print_file (const char *base, const char *name, const char *prefix,
	int *first)
{
	static int first_set = 1;

	if (par_verbose == 0) {
		return;
	}

	if (first_set) {
		first_set = 0;
	}
	else {
		if (*first) {
			fputs ("\n", stdout);
		}
	}

	if (*first) {
		printf (">%s\n", base);
	}

	printf ("%s%s\n", prefix, name);

	fflush (stdout);

	*first = 0;
}


static
void fent_init (file_entry_t *ent, char *fname)
{
	ent->name = fname;
	ent->size = 0;
	ent->chk = 0;
	ent->set = 0;
}

static
void fent_free (file_entry_t *ent)
{
	free (ent->name);
}

static
int fent_stat (file_entry_t *ent)
{
	struct stat st;

	if (lstat (ent->name, &st) != 0) {
		return (1);
	}

	ent->size = st.st_size;

	ent->dev = st.st_dev;
	ent->ino = st.st_ino;

	ent->uid = st.st_uid;
	ent->gid = st.st_gid;

	ent->mode = st.st_mode;

	ent->mtime = st.st_mtime;

	ent->chk = 0;

	return (0);
}

/*
 * Calculate the checksum for an entry
 */
static
int fent_get_checksum (file_entry_t *ent)
{
	unsigned n;
	crc32_t  crc;
	FILE     *fp;

	fp = fopen (ent->name, "rb");
	if (fp == NULL) {
		return (1);
	}

	crc32_copy (&crc, &par_crc);

	while (1) {
		n = fread (par_buf, 1, RELINK_BUF_SIZE, fp);

		if (n == 0) {
			break;
		}

		crc32_calc (&crc, par_buf, n);
	}

	if (ferror (fp)) {
		fclose (fp);
		return (1);
	}

	fclose (fp);

	crc32_done (&crc);

	ent->chk = crc32_get_crc (&crc);

	if (ent->chk == 0) {
		ent->chk = 1;
	}

	return (0);
}

/*
 * Check if two entries have identical file data
 */
static
int fent_equal_content (file_entry_t *ent1, file_entry_t *ent2)
{
	FILE          *fp1, *fp2;
	unsigned char *buf1, *buf2;
	unsigned      cnt1, cnt2;

	if ((ent1->size == 0) && (ent2->size == 0)) {
		return (1);
	}

	if (ent1->chk == 0) {
		if (fent_get_checksum (ent1)) {
			return (0);
		}

		par_chk_cnt += 1;
	}

	if (ent2->chk == 0) {
		if (fent_get_checksum (ent2)) {
			return (0);
		}

		par_chk_cnt += 1;
	}

	if ((ent1->chk != 0) && (ent2->chk != 0)) {
		if (ent1->chk != ent2->chk) {
			return (0);
		}
	}

	fp1 = fopen (ent1->name, "rb");
	if (fp1 == NULL) {
		return (0);
	}

	fp2 = fopen (ent2->name, "rb");
	if (fp2 == NULL) {
		fclose (fp1);
		return (0);
	}

	buf1 = par_buf;
	buf2 = par_buf + RELINK_BUF_SIZE / 2;

	while (1) {
		cnt1 = fread (buf1, 1, RELINK_BUF_SIZE / 2, fp1);
		cnt2 = fread (buf2, 1, RELINK_BUF_SIZE / 2, fp2);

		if (cnt1 != cnt2) {
			break;
		}

		if (cnt1 == 0) {
			break;
		}

		if (memcmp (buf1, buf2, cnt1) != 0) {
			break;
		}
	}

	fclose (fp2);
	fclose (fp1);

	par_cmp_cnt += 1;

	if ((cnt1 != 0) || (cnt2 != 0)) {
		return (0);
	}

	return (1);
}

/*
 * Check if two entries are identical according to the options chosen
 */
static
int fent_equal (file_entry_t *ent1, file_entry_t *ent2)
{
	if (ent1->dev == ent2->dev) {
		if (ent1->ino == ent2->ino) {
			return (1);
		}
	}
	else {
		return (0);
	}

	if (ent1->size != ent2->size) {
		return (0);
	}

	if (par_use_mode) {
		if (ent1->mode != ent2->mode) {
			return (0);
		}
	}

	if (par_use_mtime) {
		if (ent1->mtime != ent2->mtime) {
			return (0);
		}
	}

	if (par_use_uid) {
		if (ent1->uid != ent2->uid) {
			return (0);
		}
	}

	if (par_use_gid) {
		if (ent1->gid != ent2->gid) {
			return (0);
		}
	}

	if (fent_equal_content (ent1, ent2) == 0) {
		return (0);
	}

	return (1);
}

/*
 * Check if two entries refer to the same file
 */
static
int fent_same (file_entry_t *ent1, file_entry_t *ent2)
{
	if (ent1->dev != ent2->dev) {
		return (0);
	}

	if (ent1->ino != ent2->ino) {
		return (0);
	}

	return (1);
}

/*
 * Check if an entry is a regular file
 */
static
int fent_is_regular (file_entry_t *ent)
{
	return (S_ISREG (ent->mode) != 0);
}

static
int fent_relink (file_entry_t *dst, file_entry_t *src)
{
	if (dst->dev != src->dev) {
		return (1);
	}

	if (unlink (src->name) != 0) {
		fprintf (stderr, "%s: unlink failed (%s)\n", argv0, src->name);
		return (1);
	}

	if (link (dst->name, src->name) != 0) {
		fprintf (stderr, "%s: link failed (%s <- %s)\n", argv0,
			dst->name, src->name
		);
		return (1);
	}

	return (0);
}


static
void fgrp_init (file_group_t *grp)
{
	grp->cnt = 0;
	grp->max = 0;
	grp->ent = NULL;
}

static
void fgrp_free (file_group_t *grp)
{
	if (grp->ent != NULL) {
		free (grp->ent);
	}
}

static
void fgrp_set_done (file_group_t *grp, int val)
{
	unsigned i;

	for (i = 0; i < grp->cnt; i++) {
		grp->ent[i]->done = val;
	}
}

/*
 * Add an entry to a file group
 */
static
int fgrp_add (file_group_t *grp, file_entry_t *ent)
{
	file_entry_t **tmp;
	unsigned     max;

	if (grp->cnt >= grp->max) {
		max = grp->max;
		while (grp->cnt >= max) {
			max = (max == 0) ? 16 : (2 * max);
		}

		tmp = realloc (grp->ent, max * sizeof (file_entry_t *));
		if (tmp == NULL) {
			return (1);
		}

		grp->ent = tmp;
		grp->max = max;
	}

	grp->ent[grp->cnt] = ent;
	grp->cnt += 1;

	return (0);
}

static
int fgrp_add_group (file_group_t *dst, file_group_t *src)
{
	unsigned i;

	for (i = 0; i < src->cnt; i++) {
		if (fgrp_add (dst, src->ent[i])) {
			return (1);
		}
	}

	return (0);
}

/*
 * Get the group of all files that are linkable or linked to file idx from lst
 */
static
int fgrp_get (file_group_t *grp, file_list_t *lst, unsigned idx)
{
	unsigned     i;
	file_entry_t *ent1, *ent2;

	grp->cnt = 0;

	ent1 = &lst->ent[idx];

	if (fgrp_add (grp, ent1)) {
		return (1);
	}

	i = idx + 1;
	while (i < lst->cnt) {
		ent2 = &lst->ent[i];

		i += 1;

		if (ent1->size != ent2->size) {
			return (0);
		}

		if (ent2->done) {
			continue;
		}

		if (fent_equal (ent1, ent2) == 0) {
			continue;
		}

		if (fgrp_add (grp, ent2)) {
			return (1);
		}
	}

	return (0);
}

/*
 * Add the subgroup of files that are linked to ent from src to dst
 */
static
void fgrp_get_same (file_group_t *dst, file_group_t *src, file_entry_t *ent)
{
	unsigned i;

	for (i = 0; i < src->cnt; i++) {
		if (src->ent[i]->done) {
			continue;
		}

		if (fent_same (ent, src->ent[i])) {
			fgrp_add (dst, src->ent[i]);
		}
	}
}

/*
 * Check if any entry in grp1 is in the same set as any entry in grp2
 */
static
int fgrp_check_set (const file_group_t *grp1, const file_group_t *grp2)
{
	unsigned i, j;

	for (i = 0; i < grp1->cnt; i++) {
		for (j = 0; j < grp2->cnt; j++) {
			if (grp1->ent[i]->set == grp2->ent[j]->set) {
				return (1);
			}
		}
	}

	return (0);
}

static
int fgrp_sort_cmp (const void *el1, const void *el2)
{
	const file_entry_t *ent1, *ent2;

	ent1 = *(const file_entry_t **) el1;
	ent2 = *(const file_entry_t **) el2;

	if (ent1->mtime < ent2->mtime) {
		return (par_reverse ? 1 : -1);
	}
	else if (ent1->mtime > ent2->mtime) {
		return (par_reverse ? -1 : 1);
	}

	return (strcmp (ent1->name, ent2->name));
}

static
void fgrp_sort (file_group_t *grp)
{
	qsort (grp->ent, grp->cnt, sizeof (file_entry_t *), fgrp_sort_cmp);
}

static
int fgrp_relink (file_group_t *grp)
{
	unsigned     i;
	int          first;
	file_entry_t *ent1, *ent2;

	if (grp->cnt < 2) {
		return (0);
	}

	fgrp_sort (grp);

	first = 1;

	ent1 = grp->ent[0];

	for (i = 1; i < grp->cnt; i++) {
		ent2 = grp->ent[i];

		if (fent_same (ent1, ent2)) {
			if (par_print_same) {
				relink_print_file (
					ent1->name, ent2->name, "=", &first);
			}
		}
		else {
			relink_print_file (ent1->name, ent2->name, "+", &first);

			if (par_link) {
				if (fent_relink (ent1, ent2)) {
					;
				}
			}

			par_lnk_cnt += 1;
		}

	}

	return (0);
}

static
int fgrp_relink_diff (file_group_t *grp)
{
	unsigned     i, j;
	file_group_t sub1, sub2;

	if (grp->cnt < 2) {
		return (0);
	}

	fgrp_sort (grp);

	fgrp_init (&sub1);
	fgrp_init (&sub2);

	for (i = 0; i < grp->cnt; i++) {
		if (grp->ent[i]->done) {
			continue;
		}

		sub1.cnt = 0;
		fgrp_get_same (&sub1, grp, grp->ent[i]);
		fgrp_set_done (&sub1, 1);

		for (j = i + 1; j < grp->cnt; j++) {
			if (grp->ent[j]->done) {
				continue;
			}

			sub2.cnt = 0;
			fgrp_get_same (&sub2, grp, grp->ent[j]);

			if (fgrp_check_set (&sub1, &sub2) == 0) {
				fgrp_set_done (&sub2, 1);
				fgrp_add_group (&sub1, &sub2);
			}
		}

		fgrp_sort (&sub1);

		if (fgrp_relink (&sub1)) {
			return (1);
		}

		fgrp_set_done (&sub1, 1);
	}

	fgrp_free (&sub2);
	fgrp_free (&sub1);

	return (0);
}


static
void flst_init (file_list_t *lst)
{
	lst->cnt = 0;
	lst->max = 0;
	lst->ent = NULL;
}

static
void flst_free (file_list_t *lst)
{
	unsigned long i;

	for (i = 0; i < lst->cnt; i++) {
		fent_free (&lst->ent[i]);
	}

	free (lst->ent);
}

static
file_entry_t *flst_add_entry (file_list_t *lst, char *name)
{
	unsigned long cnt;
	file_entry_t  *ent;

	if (lst->cnt >= lst->max) {
		cnt = (lst->max == 0) ? 64 : lst->max;

		while (cnt <= lst->cnt) {
			cnt += cnt;
		}

		ent = realloc (lst->ent, cnt * sizeof (file_entry_t));
		if (ent == NULL) {
			free (name);
			return (NULL);
		}

		lst->ent = ent;
		lst->max = cnt;
	}

	ent = &lst->ent[lst->cnt++];

	fent_init (ent, name);

	return (ent);
}

static
int flst_remove_last_entry (file_list_t *lst)
{
	if (lst->cnt == 0) {
		return (1);
	}

	fent_free (&lst->ent[lst->cnt - 1]);

	lst->cnt -= 1;

	return (0);
}

static
int flst_sort_cmp (const void *el1, const void *el2)
{
	const file_entry_t *ent1, *ent2;

	ent1 = el1;
	ent2 = el2;

	if (ent1->size < ent2->size) {
		return (-1);
	}
	else if (ent1->size > ent2->size) {
		return (1);
	}

	return (0);
}

/*
 * Sort a file list by size
 */
static
void flst_sort (file_list_t *lst)
{
	qsort (lst->ent, lst->cnt, sizeof (file_entry_t), flst_sort_cmp);
}

static
char *flst_read_line (FILE *fp)
{
	unsigned    i;
	int         c;
	static char buf[512];

	i = 0;
	while (i < 512) {
		c = fgetc (fp);

		if (c == EOF) {
			break;
		}
		else if ((c == 0x0d) || (c == 0x0a)) {
			if (i > 0) {
				break;
			}
		}
		else {
			buf[i++] = (char) c;
		}
	}

	if ((i == 0) || (i > 511)) {
		return (NULL);
	}

	/* trailing slashes */
	while ((i > 1) && (buf[i - 1] == '/')) {
		i -= 1;
	}

	buf[i] = 0;

	/* leading "./" */
	i = 0;
	while ((buf[i] == '.') && (buf[i + 1] == '/')) {
		i += 2;
	}

	if (buf[i] == 0) {
		buf[0] = '.';
		buf[1] = 0;
		i = 0;
	}

	return (strdup (buf + i));
}

/*
 * Read a file list from fp and add it to lst
 */
static
int flst_read_fp (file_list_t *lst, FILE *fp, unsigned set)
{
	char         *fname;
	file_entry_t *ent;

	if (par_progress) {
		fprintf (stderr, "reading... %lu\r", lst->cnt);
		fflush (stderr);
	}

	while (1) {
		fname = flst_read_line (fp);

		if (fname == NULL) {
			if (par_progress) {
				fputs ("\n", stderr);
				fflush (stderr);
			}

			if (ferror (fp)) {
				return (1);
			}

			return (0);
		}

		ent = flst_add_entry (lst, fname);

		if (ent == NULL) {
			return (1);
		}

		ent->set = set;

		if (fent_stat (ent)) {
			fprintf (stderr, "%s: skipping %s\n",
				argv0, ent->name
			);
			flst_remove_last_entry (lst);
		}
		else if (fent_is_regular (ent) == 0) {
			flst_remove_last_entry (lst);
		}
		else {
			if (par_progress) {
				fprintf (stderr, "reading... %lu\r", lst->cnt);
				fflush (stderr);
			}
		}
	}

	return (1);
}

/*
 * Read a file list from name and add it to lst
 */
static
int flst_read (file_list_t *lst, const char *name, unsigned set)
{
	int  r;
	FILE *fp;

	if ((name == NULL) || (strcmp (name, "-") == 0)) {
		fp = stdin;
	}
	else {
		fp = fopen (name, "r");
		if (fp == NULL) {
			fprintf (stderr, "%s: can't open file (%s)\n",
				argv0, name
			);
			return (1);
		}
	}

	r = flst_read_fp (lst, fp, set);

	if (fp != stdin) {
		fclose (fp);
	}

	return (r);
}

static
int flst_relink (file_list_t *lst)
{
	unsigned long i;
	file_group_t  grp;

	if (par_progress) {
		fprintf (stderr, "sorting...\n");
		fflush (stderr);
	}

	flst_sort (lst);

	for (i = 0; i < lst->cnt; i++) {
		lst->ent[i].done = 0;
	}

	fgrp_init (&grp);

	for (i = 0; i < lst->cnt; i++) {
		if (par_progress) {
			fprintf (stderr,
				"[%6.2f%%] %lu+%lu=%lu CHK=%lu LNK=%lu \r",
				100.0 * (double) (i + 1) / (double) lst->cnt,
				(i + 1), lst->cnt - i - 1, lst->cnt,
				par_chk_cnt, par_lnk_cnt
			);
			fflush (stderr);
		}

		if (lst->ent[i].done) {
			continue;
		}

		if (fgrp_get (&grp, lst, i)) {
			return (1);
		}

		if (par_set_diff) {
			if (fgrp_relink_diff (&grp)) {
				return (1);
			}
		}
		else {
			if (fgrp_relink (&grp)) {
				return (1);
			}
		}

		fgrp_set_done (&grp, 1);
	}

	fgrp_free (&grp);

	if (par_progress) {
		fputs ("\n", stderr);
		fflush (stderr);
	}

	return (0);
}

int main (int argc, char **argv)
{
	int         i, r;
	int         ok;
	file_list_t lst;
	unsigned    set;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (0);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("relink");
			return (0);
		}
	}

	flst_init (&lst);

	set = 0;
	ok = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "g", "no-gid")) {
			par_use_gid = 0;
		}
		else if (str_isarg2 (argv[i], "i", "input")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input\n", argv0);
				return (1);
			}

			if (flst_read (&lst, argv[i], set)) {
				return (1);
			}

			ok = 1;
		}
		else if (str_isarg2 (argv[i], "l", "link")) {
			par_link = 1;
		}
		else if (str_isarg2 (argv[i], "m", "no-mode")) {
			par_use_mode = 0;
		}
		else if (str_isarg2 (argv[i], "n", "next-set")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input\n", argv0);
				return (1);
			}

			set += 1;

			if (flst_read (&lst, argv[i], set)) {
				return (1);
			}

			par_set_diff = 1;
			ok = 1;
		}
		else if (str_isarg2 (argv[i], "p", "progress")) {
			par_progress = 1;
		}
		else if (str_isarg2 (argv[i], "r", "reverse")) {
			par_reverse = 1;
		}
		else if (str_isarg2 (argv[i], "s", "print-same")) {
			par_print_same = 1;
		}
		else if (str_isarg2 (argv[i], "t", "no-mtime")) {
			par_use_mtime = 0;
		}
		else if (str_isarg2 (argv[i], "u", "no-uid")) {
			par_use_uid = 0;
		}
		else if (str_isarg2 (argv[i], "v", "verbose")) {
			par_verbose = 1;
		}
		else if (argv[i][0] == '-') {
			fprintf (stderr, "%s: unknown option (%s)\n",
				argv[0], argv[i]
			);
			return (1);
		}
		else {
			if (flst_read (&lst, argv[i], set)) {
				return (1);
			}

			ok = 1;
		}

		i += 1;
	}

	if (ok == 0) {
		if (flst_read_fp (&lst, stdin, set)) {
			return (1);
		}
	}

	crc32_init (&par_crc, CRC_CRC32, CRC_POSIX);

	if (lst.cnt > 0) {
		r = flst_relink (&lst);
	}
	else {
		r = 0;
	}

	flst_free (&lst);

	if (r) {
		return (1);
	}

	return (0);
}
