/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     rpipe.c                                                    *
 * Created:       2002-01-17 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-11-19 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2002-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: rpipe.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/common.h>
#include <lib/socket.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "rpipe.h"


#define RPIPE_BUF_SIZE MISC_BUF_SIZE


static const char     *argv0 = NULL;

static int            par_verbose = 0;
static char           *par_host = NULL;
static char           *par_addr = NULL;
static unsigned short par_port = 0;
static int            par_input = 0;
static int            par_term = 0;
static unsigned       par_retry_delay = 2;
static unsigned long  par_retry_cnt = 0;

static int            par_reading = 0;
static int            par_writing = 0;


static
void prt_help (void)
{
	fputs (
		"usage: rpipe [options]\n"
		"  --help                   Print usage information\n"
		"  --version                Print version information\n"
		"  -v, --verbose            Verbose operation [no]\n"
		"  -i, --input              Listen to a port (default if -p is used)\n"
		"  -o, --output             Connect to a port (default if -h is used)\n"
		"  -h, --host string[:int]  Set host name and port\n"
		"  -p, --port int           Set port\n"
		"  -r, --read               Read from pipe (default if input)\n"
		"  -w, --write              Write to pipe (default if output)\n"
		"  -n, --retry-count int    Set the retry count [0]\n"
		"  -t, --retry-delay int    Set the retry delay in seconds [2]\n",
		stdout
	);

	fflush (stdout);
}

static
void sig_term (int sig)
{
	par_term = 1;
}

static
void sig_pipe (int sig)
{
}

static
int rpipe_copy (int src, int dst)
{
	ssize_t       cnt;
	unsigned      buf_i, buf_n;
	unsigned char buf[RPIPE_BUF_SIZE];


	cnt = read (src, buf, RPIPE_BUF_SIZE);
	if (par_term || (cnt < 0)) {
		return (1);
	}

	while (cnt > 0) {
		buf_i = 0;
		buf_n = (unsigned) cnt;

		while (buf_n > 0) {
			cnt = write (dst, buf + buf_i, buf_n);
			if (par_term || (cnt <= 0)) {
				return (1);
			}

			buf_i += (unsigned) cnt;
			buf_n -= (unsigned) cnt;
		}

		cnt = read (src, buf, RPIPE_BUF_SIZE);
		if (par_term || (cnt < 0)) {
			return (1);
		}
	}

	return (0);
}

static
int rpipe_handle (int fd)
{
	int   r;
	int   status;
	pid_t pid, tmp;

	if (par_reading && par_writing) {
		pid = fork();
		if (pid < 0) {
			fprintf (stderr, "%s: fork failed\n", argv0);
			return (1);
		}

		if (pid == 0) {
			r = rpipe_copy (fd, 1);
			tcp_shutdown_inp (fd);
			exit (r);
		}

		r = rpipe_copy (0, fd);
		tcp_shutdown_out (fd);

		while (par_term == 0) {
			tmp = waitpid (pid, &status, 0);
			if (tmp < 0) {
				return (1);
			}

			if (tmp == pid) {
				return (r || status);
			}
		}

		return (r);
	}

	if (par_reading) {
		return (rpipe_copy (fd, 1));
	}

	if (par_writing) {
		return (rpipe_copy (0, fd));
	}

	return (1);
}

static
int rpipe_inp (void)
{
	int r;
	int fd1, fd2;

	fd1 = tcp_listen (par_addr, par_port, 16);
	if (fd1 == -1) {
		fprintf (stderr, "%s: can't create socket for port %u\n", argv0, par_port);
		return (1);
	}

	fd2 = -1;

	while ((fd2 < 0) && (par_term == 0)) {
		fd2 = tcp_accept_time (fd1, 500);
	}

	close (fd1);

	if (fd2 < 0) {
		return (1);
	}

	if (par_verbose) {
		char           *host;
		unsigned short port;

		if (tcp_get_peer (fd2, &host, &port) == 0) {
			fprintf (stderr, "rpipe: accepted connection from %s:%u\n", host, port);
			fflush (stderr);
			free (host);
		}
	}

	r = rpipe_handle (fd2);

	close (fd2);

	return (r);
}

static
int rpipe_out (void)
{
	int r;
	int fd;

	if (par_addr == NULL) {
		fprintf (stderr, "%s: need a host name\n", argv0);
		return (1);
	}

	fd = -1;

	while (fd < 0) {
		if (par_term) {
			return (1);
		}

		fd = tcp_connect (par_addr, par_port);

		if (fd < 0) {
			unsigned tm;

			if (par_retry_cnt > 0) {
				par_retry_cnt -= 1;

				if (par_retry_cnt == 0) {
					fprintf (stderr, "%s: exceeded retry count\n", argv0);
					return (1);
				}
			}

			tm = par_retry_delay;
			while (tm > 0) {
				if (par_term) {
					return (1);
				}

				tm = sleep (tm);
			}
		}
	}

	if (par_verbose) {
		fprintf (stderr, "rpipe: connected to %s:%u\n", par_addr, (unsigned) par_port);
		fflush (stderr);
	}

	r = rpipe_handle (fd);

	close (fd);

	return (r);
}

int main (int argc, char *argv[])
{
	int            i, r;
	unsigned short port;
	int            have_dir;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("rpipe");
			return (EXIT_SUCCESS);
		}
	}

	have_dir = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "i", "input")) {
			par_input = 1;
			have_dir = 1;
		}
		else if (str_isarg2 (argv[i], "o", "output")) {
			par_input = 0;
			have_dir = 1;
		}
		else if (str_isarg2 (argv[i], "h", "host")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing host name\n", argv0);
				return (EXIT_FAILURE);
			}

			par_host = tcp_addr_get_host (argv[i]);

			if (tcp_addr_get_port (argv[i], &port) == 0) {
				par_port = port;
			}

			if (!have_dir) {
				par_input = 0;
				have_dir = 1;
			}
		}
		else if (str_isarg2 (argv[i], "p", "port")) {
			unsigned tmp;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing port number\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &tmp)) {
				fprintf (stderr, "%s: bad port number (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_port = tmp;

			if (!have_dir) {
				par_input = 1;
				have_dir = 1;
			}
		}
		else if (str_isarg2 (argv[i], "r", "read")) {
			par_reading = 1;
		}
		else if (str_isarg2 (argv[i], "w", "write")) {
			par_writing = 1;
		}
		else if (str_isarg2 (argv[i], "n", "retry-count")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing retry count\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_retry_cnt)) {
				fprintf (stderr, "%s: bad retry count (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "t", "retry-dely")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing retry delay\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_retry_delay)) {
				fprintf (stderr, "%s: bad retry delay (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "v", "verbose")) {
			par_verbose = 1;
		}
		else {
			fprintf (stderr, "%s: unknown parameter (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (par_host != NULL) {
		par_addr = tcp_host_resolve (par_host);
	}

	if ((par_reading == 0) && (par_writing == 0)) {
		if (par_input) {
			par_reading = 1;
		}
		else {
			par_writing = 1;
		}
	}

	signal (SIGINT, &sig_term);
	signal (SIGTERM, &sig_term);
	signal (SIGPIPE, &sig_pipe);

	if (par_input) {
		r = rpipe_inp();
	}
	else {
		r = rpipe_out();
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
