/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     sha256.c                                                   *
 * Created:       2005-07-27 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2006-06-01 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: sha256.c 390 2006-06-01 09:03:57Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>
#include <lib/sha256.h>

#include "sha256.h"


#define SHA256_BUF_SIZE MISC_BUF_SIZE


static const char *argv0 = NULL;

static int      par_enc = 0;
static int      par_dec = 0;
static int      par_bin = 0;
static int      par_pipe = 0;
static int      par_capital = 0;
static unsigned par_digits = 0;


static
void prt_help (void)
{
	fputs (
		"Usage: sha256 [options] [input [output]]\n"
		"  -i string         Set input file name [stdin]\n"
		"  -o string         Set output file name [stdout]\n"
		"  -e, --encode      Add binary SHA-256 sum to output [no]\n"
		"  -d, --decode      Verify and remove binary SHA-256 sum from input [no]\n"
		"  -b, --binary      Write the SHA-256 sum as binary string [no]\n"
		"  -n, --digits int  Group n digits in output [0]\n"
		"  -c, --capital     Use capital letters as hex digits [no]\n"
		"  -p, --pipe        Pass input through to stdout [no]\n",
		stdout
	);

	fflush (stdout);
}

static
int do_sha256_enc (FILE *inp, FILE *out)
{
	unsigned long n;
	sha256_t      sha;
	unsigned char buf[SHA256_BUF_SIZE + 32];

	sha256_init (&sha);

	n = fread (buf, 1, SHA256_BUF_SIZE, inp);

	while (n > 0) {
		sha256_calc (&sha, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, SHA256_BUF_SIZE, inp);
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	sha256_done (&sha);

	sha256_get_sum (&sha, buf, 32);

	if (fwrite (buf, 1, 32, out) != 32) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}

static
int do_sha256_dec (FILE *inp, FILE *out)
{
	unsigned      i;
	unsigned long n;
	sha256_t      sha;
	unsigned char buf[SHA256_BUF_SIZE + 64];

	sha256_init (&sha);

	n = fread (buf, 1, SHA256_BUF_SIZE + 32, inp);

	while (n > 32) {
		sha256_calc (&sha, buf, n - 32);

		if (fwrite (buf, 1, n - 32, out) != (n - 32)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		for (i = 0; i < 32; i++) {
			buf[i] = buf[n - 32 + i];
		}

		n = 32 + fread (buf + 32, 1, SHA256_BUF_SIZE, inp);
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	sha256_done (&sha);

	sha256_get_sum (&sha, buf + 32, 32);

	for (i = 0; i < 32; i++) {
		if (buf[i] != buf[i + 32]) {
			fprintf (stderr, "%s: sha256 sum mismatch\n", argv0);
			return (1);
		}
	}

	return (0);
}

static
char sha256_get_hex (unsigned char val)
{
	if (val < 10) {
		return ((char) ('0' + val));
	}

	if (par_capital) {
		return ((char) ('A' + val - 10));
	}

	return ((char) ('a' + val - 10));
}

static
int do_sha256 (FILE *inp, FILE *out)
{
	unsigned      i, j;
	unsigned long n;
	sha256_t      sha;
	unsigned char buf[SHA256_BUF_SIZE + 32];

	sha256_init (&sha);

	n = fread (buf, 1, SHA256_BUF_SIZE, inp);

	while (n > 0) {
		sha256_calc (&sha, buf, n);

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, SHA256_BUF_SIZE, inp);
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	sha256_done (&sha);

	sha256_get_sum (&sha, buf, 32);

	if (par_bin) {
		if (fwrite (buf, 1, 32, out) != 32) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}
	}
	else {
		j = 0;
		for (i = 0; i < 32; i++) {
			fputc (sha256_get_hex (buf[i] >> 4), out);
			j += 1;
			if ((par_digits > 0) && (j >= par_digits)) {
				fputc (' ', out);
				j = 0;
			}

			fputc (sha256_get_hex (buf[i] & 0x0f), out);
			j += 1;
			if ((par_digits > 0) && (j >= par_digits)) {
				fputc (' ', out);
				j = 0;
			}
		}

		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}

int main (int argc, char *argv[])
{
	int        i, r;
	const char *finp, *fout;
	FILE       *inp, *out;
	unsigned   fcnt;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("sha256");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "e", "encode")) {
			par_enc = 1;
			par_dec = 0;
		}
		else if (str_isarg2 (argv[i], "d", "decode")) {
			par_enc = 0;
			par_dec = 1;
		}
		else if (str_isarg2 (argv[i], "b", "binary")) {
			par_bin = 1;
		}
		else if (str_isarg2 (argv[i], "n", "digits")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing number of digits\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_uint (argv[i], &par_digits)) {
				fprintf (stderr, "%s: bad number of digits (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "c", "capital")) {
			par_capital = 1;
		}
		else if (str_isarg2 (argv[i], "p", "pipe")) {
			par_pipe = 1;
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown argument (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	inp = (finp == NULL) ? stdin : fopen (finp, "rb");
	if (inp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
		return (EXIT_FAILURE);
	}

	out = (fout == NULL) ? stdout : fopen (fout, par_bin ? "wb" : "w");
	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
		return (EXIT_FAILURE);
	}

	if (par_enc) {
		r = do_sha256_enc (inp, out);
	}
	else if (par_dec) {
		r = do_sha256_dec (inp, out);
	}
	else {
		r = do_sha256 (inp, out);
	}

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
