/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     sparse.c                                                   *
 * Created:       2003-07-09 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-11-19 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2003-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: sparse.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/config.h>
#include <lib/common.h>
#include <lib/file.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <stdlib.h>

#ifdef HAVE_FTRUNCATE
#include <unistd.h>
#include <sys/types.h>
#endif

#include "sparse.h"


#define SPARSE_BUF_SIZE MISC_BUF_SIZE


static const char *argv0 = NULL;

#ifdef HAVE_FSEEKO
static off_t par_zero_cnt = 0;
static off_t par_size = 0;
#else
static long par_zero_cnt = 0;
static long par_size = 0;
#endif

static int   par_create = 0;


static
void prt_help (void)
{
	fputs (
		"Usage: sparse [options] [input [output]]\n"
		"  --help            Print usage information\n"
		"  --version         Print version information\n"
		"  -i string         Set input file name [stdin]\n"
		"  -o string         Set output file name [stdout]\n"
		"  -c, --create int  Create an empty sparse file\n",
		stdout
	);

	fflush (stdout);
}

static
int sparse_write (FILE *fp, const void *buf, unsigned cnt)
{
	unsigned            i;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	i = 0;
	while ((i < cnt) && (tmp[i] == 0)) {
		par_zero_cnt += 1;
		i += 1;
	}

	if (i >= cnt) {
		return (0);
	}

	if (par_zero_cnt > 0) {
#ifdef HAVE_FSEEKO
		if (fseeko (fp, par_zero_cnt, SEEK_CUR)) {
			return (1);
		}
#else
		if (fseek (fp, par_zero_cnt, SEEK_CUR)) {
			return (1);
		}
#endif

		par_zero_cnt = 0;
	}

	if (fwrite (tmp + i, cnt - i, 1, fp) != 1) {
		return (1);
	}

	return (0);
}

static
int sparse_flush (FILE *fp)
{
	if (par_zero_cnt == 0) {
		return (0);
	}

#ifdef HAVE_FSEEKO
	if (fseeko (fp, par_zero_cnt - 1, SEEK_CUR)) {
		return (1);
	}
#else
	if (fseek (fp, par_zero_cnt - 1, SEEK_CUR)) {
		return (1);
	}
#endif

	par_zero_cnt = 0;

	fputc (0, fp);

	fflush (fp);

	return (0);
}

static
int sparse_make (FILE *inp, FILE *out)
{
	size_t        r;
	unsigned char buf[SPARSE_BUF_SIZE];

	while (1) {
		r = fread (buf, 1, SPARSE_BUF_SIZE, inp);

		if (r < 0) {
			return (1);
		}

		if (r == 0) {
			break;
		}

		if (sparse_write (out, buf, r)) {
			return (1);
		}

		par_size += r;
	}

	if (sparse_flush (out)) {
		return (1);
	}

	return (0);
}

#ifdef HAVE_FTRUNCATE
static
int sparse_create (FILE *fp)
{
	int fd;

	fd = fileno (fp);

	if (ftruncate (fd, par_size)) {
		return (1);
	}

	return (0);
}
#else
static
int sparse_create (FILE *fp)
{
	par_zero_cnt = par_size;

	if (sparse_flush (fp)) {
		return (1);
	}

	return (0);
}
#endif

static
int str_get_ullng (const char *str, unsigned long long *ret)
{
	unsigned long long val;
	char               *end;

	val = strtoull (str, &end, 0);

	if (end == str) {
		return (1);
	}

	if ((*end == 'k') || (*end == 'K')) {
		val *= 1024;
	}
	else if ((*end == 'm') || (*end == 'M')) {
		val *= 1024;
		val *= 1024;
	}
	else if ((*end == 'g') || (*end == 'G')) {
		val *= 1024;
		val *= 1024;
		val *= 1024;
	}

	*ret = val;

	return (0);
}

int main (int argc, char *argv[])
{
	int        i, r;
	unsigned   fcnt;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("sparse");
			return (EXIT_SUCCESS);
		}
	}

	finp = NULL;
	fout = NULL;
	fcnt = 0;

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			finp = argv[i];
			fcnt = 2;
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			fout = argv[i];
			fcnt = 2;
		}
		else if (str_isarg2 (argv[i], "c", "create")) {
			unsigned long long tmp;

			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing file size\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ullng (argv[i], &tmp)) {
				fprintf (stderr, "%s: bad file size (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_size = (off_t) tmp;

			if ((par_size < 0) || ((unsigned long long) par_size != tmp)) {
				fprintf (stderr, "%s: file size too large (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}

			par_create = 1;
		}
		else if (argv[i][0] != '-') {
			if (fcnt == 0) {
				finp = argv[i];
				fcnt = 1;
			}
			else if (fcnt == 1) {
				fout = argv[i];
				fcnt = 2;
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, "rb");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, "wb");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
			return (EXIT_FAILURE);
		}
	}

	if (par_create) {
		r = sparse_create (out);
	}
	else {
		r = sparse_make (inp, out);
	}

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
