/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     ttylog.c                                                   *
 * Created:       2002-02-04 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-07-10 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2002-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: ttylog.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/common.h>
#include <lib/file.h>
#include <lib/strarg.h>
#include <lib/strtoint.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "ttylog.h"


static const char    *argv0 = NULL;

static int           par_decode = 0;
static int           par_pipe = 0;

static unsigned long par_scale = 100;
static unsigned long par_max = 5000;
static unsigned long par_threshold = 2;

static file_t        par_fpipe;


static
void prt_help (void)
{
	fputs (
		"Usage: ttylog [options]\n"
		"\nOptions are:\n"
		"  --help               Print usage information\n"
		"  --version            Print version information\n"
		"  -i string            Set input file name [stdin]\n"
		"  -o string            Set output file name [stdout]\n"
		"  -e, --encode         Encode a character stream\n"
		"  -d, --decode         Decode a character stream\n"
		"  -m, --max int        Set maximum delay in milliseconds [5000]\n"
		"  -s, --scale int      Set time scale in percent [100]\n"
		"  -t, --threshold int  Set the threshold delay [2]\n"
		"  -p, --pipe           Copy input unchanged to stdout\n",
		stdout
	);

	fflush (stdout);
}

static
unsigned long get_delta_time (struct timeval *tv)
{
	unsigned long  ret;
	long           tmp;
	struct timeval tvnew;

	gettimeofday (&tvnew, NULL);

	ret = 1000 * (unsigned long) (tvnew.tv_sec - tv->tv_sec);
	tmp = (tvnew.tv_usec - tv->tv_usec) / 1000;

	if (tmp < 0) {
		ret -= 1000;
		tmp += 1000;
	}

	ret += (unsigned long) tmp;

	if (ret <= par_threshold) {
		ret = 0;
	}

	*tv = tvnew;

	return (ret);
}

static
void set_delta_time (unsigned long tm)
{
	struct timeval tv;

	tm = (par_scale * tm) / 100;

	if (tm > par_max) {
		tm = par_max;
	}

	if (tm > 0) {
		tv.tv_sec = (long) (tm / 1000);
		tv.tv_usec = (long) (1000 * (tm % 1000));

		select (0, NULL, NULL, NULL, &tv);
	}
}

static
int do_encode (file_t *f_inp, file_t *f_out)
{
	int            c;
	unsigned long  tm;
	unsigned char  buf[5];
	struct timeval tv;

	gettimeofday (&tv, NULL);
	c = file_get_char (f_inp);

	while (c != EOF) {
		tm = get_delta_time (&tv);

		buf[0] = (tm >> 24) & 0xff;
		buf[1] = (tm >> 16) & 0xff;
		buf[2] = (tm >> 8) & 0xff;
		buf[3] = tm & 0xff;
		buf[4] = c & 0xff;

		file_write_block (f_out, buf, 5);
		file_flush (f_out);

		if (par_pipe) {
			file_set_char (&par_fpipe, c);
			file_flush (&par_fpipe);
		}

		c = file_get_char (f_inp);
	}

	return (0);
}

static
int do_decode (file_t *f_inp, file_t *f_out)
{
	int           c;
	unsigned long r;
	unsigned long tm;
	unsigned char buf[5];

	r = file_read_block (f_inp, buf, 5);

	while (r == 5) {
		tm = buf[0] & 0xff;
		tm = (tm << 8) + (buf[1] & 0xff);
		tm = (tm << 8) + (buf[2] & 0xff);
		tm = (tm << 8) + (buf[3] & 0xff);
		c = buf[4];

		if (tm > par_threshold) {
			set_delta_time (tm);
		}

		file_set_char (f_out, c);
		file_flush (f_out);

		if (par_pipe) {
			file_write_block (&par_fpipe, buf, 5);
			file_flush (&par_fpipe);
		}

		r = file_read_block (f_inp, buf, 5);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int    i, r;
	file_t f_inp;
	file_t f_out;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help ();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("ttylog");
			return (EXIT_SUCCESS);
		}
	}

	file_set_stdin (&f_inp);
	file_set_stdout (&f_out);

	i = 1;
	while (i < argc) {
		if (str_isarg1 (argv[i], "i")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing input file name\n", argv0);
				return (EXIT_FAILURE);
			}

			file_set_name (&f_inp, argv[i]);
		}
		else if (str_isarg1 (argv[i], "o")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing output file name\n", argv0);
				return (EXIT_FAILURE);
			}

			file_set_name (&f_out, argv[i]);
		}
		else if (str_isarg2 (argv[i], "e", "encode")) {
			par_decode = 0;
		}
		else if (str_isarg2 (argv[i], "d", "decode")) {
			par_decode = 1;
		}
		else if (str_isarg2 (argv[i], "p", "pipe")) {
			par_pipe = 1;
			file_set_stdout (&par_fpipe);
		}
		else if (str_isarg2 (argv[i], "m", "max")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing maximum delta time\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_max)) {
				fprintf (stderr, "%s: bad maximum delta time (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "s", "scale")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing time scale\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_scale)) {
				fprintf (stderr, "%s: bad time scale (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else if (str_isarg2 (argv[i], "t", "threshold")) {
			i += 1;
			if (i >= argc) {
				fprintf (stderr, "%s: missing threshold value\n", argv0);
				return (EXIT_FAILURE);
			}

			if (str_get_ulng (argv[i], &par_threshold)) {
				fprintf (stderr, "%s: bad threshold (%s)\n", argv0, argv[i]);
				return (EXIT_FAILURE);
			}
		}
		else {
			fprintf (stderr, "%s: unknown option (%s)\n", argv0, argv[i]);
			return (EXIT_FAILURE);
		}

		i += 1;
	}

	if (file_open_inp (&f_inp)) {
		fprintf (stderr, "%s: can't open input file\n", argv0);
		return (EXIT_FAILURE);
	}

	if (file_open_out (&f_out)) {
		fprintf (stderr, "%s: can't open output file\n", argv0);
		return (EXIT_FAILURE);
	}

	if (par_decode == 0) {
		r = do_encode (&f_inp, &f_out);
	}
	else {
		r = do_decode (&f_inp, &f_out);
	}

	file_close (&f_out);
	file_close (&f_inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
